/*******************************************************************
HILBEX
version 0.1, 1 Oct 2016

This file is part of the software package Hilbex by Herman
Haverkort. Use and distribution of this software is subject to
restrictions. Consult readme.txt for details, or download it from
http://spacefillingcurves.net/.
*******************************************************************/

#include <iostream>
using std::ostream;
#include <fstream>
using std::ifstream;
using std::ofstream;
#include <iomanip>
using std::cin;
using std::cout;
using std::cerr;
using std::endl;
#include <vector>
using std::vector;
#include <stack>
using std::stack;
#include <string>
using std::string;
#include <sstream>
using std::stringstream;
#include <stdlib.h>
#include <time.h>

int hexbit(const int x, const int i) { return (x >> (3-i) & 1); }

#include "geometry.h"
#include "curve.h"
#include "search.h"
#include "filter.h"
#include "options.h"

#include "geometry.cpp"
#include "basepatterns.cpp"
#include "curve.cpp"
#include "search.cpp"
#include "properties.cpp" 
#include "filter.cpp" 
#include "options.cpp"
#include "traversal.cpp"

int main(int argc, char* argv[])
{
  cerr << "You are starting software that is provided for the purpose of allowing" << endl
       << "researchers to verify the findings in the manuscript" << endl
       << "\"How many three-dimensional Hilbert curves are there?\"" << endl
       << "by Herman Haverkort. Not all features of the tool have undergone" << endl
       << "adequate testing. RUNNING THIS TOOL MAY DAMAGE YOUR SYSTEM, for example," << endl
       << "by generating large or unbounded amounts of output. Use and distribution" << endl
       << "of this software is subject to restrictions. Consult readme.txt for details," << endl
       << "or download it from http://spacefillingcurves.net/." << endl << endl
       << "Please enter \"I understand\" to continue:" << endl << endl;
  char understood[21];
  cin.getline(understood, 20);
  if (string(understood) != string("I understand")) 
  {
    cerr << "I received the following response: \"" << understood << "\"," << endl << "so I cannot continue." << endl;
    exit(1);
  }
       
  initializeBasePatternTable();
  ifstream* curveInfo = new ifstream("curveinfo.txt");
  srand (time(NULL));

  Options options(argc - 1, argv + 1);
  int lineNr;
    
  do
  {
    Filter* filter = &options.filter;
    if (options.batched) 
    {
      char c = cin.peek();
      if (!cin.good()) break;
      ++lineNr; if (options.progressIndicator) cerr << "Line " << lineNr << "...\r";
      filter = new Filter(cin, options.filter);
    }
    Hilbert3DCurveSearch search(*filter, options.progressIndicator && !options.batched, options.basePatternVerificationNumbers);
    int nrFound = 0;
    do 
    {
      Hilbert3DCurve h(search, options.testLevel);
      if (h.is(Invalid)) break;
      ++nrFound;
      if (nrFound <= options.maxOutput)
      {
        Transformation t = options.transformation;
        if (options.randomTransformation) t = Transformation(rand() % 96);
        vector<OPt> v;
           
        bool space = false;
        if (options.outputProperties == Concise) { cout << h.propertyTable(); space = true; }
        if (options.outputNames == Concise) { if (space) cout << ' '; cout << h.shortName(); space = true; } else
        if (options.outputNames == Full) { if (space) cout << ' '; cout << h.name(); space = true; }
        if (options.outputDefinition) { if (space) cout << ' '; cout << h.description(t); space = true; }      
  
        if (options.traversalOrder > 0)
        {
          h.generateTraversal(v, options.traversalOrder, t);
          if (!options.povRayOutput)
          {
            if (space) cout << ' ';
            for (vector<OPt>::iterator i = v.begin(); i != v.end(); ++i) cout << convert(*i, options.traversalOrder);
            space = true;
          }
        }
        if (options.outputProperties == Full) 
          for (int i = 0; i < (int) Invalid; ++i)
            if (h.is((Property) i)) { if (space) cout << ' '; cout << Properties::name[i]; space = true; }
        if (!options.povRayOutput && (
          options.outputProperties != None || 
          options.outputNames != None || 
          options.outputDefinition != None || 
          options.traversalOrder > 0
        )) cout << endl;
        if (options.outputNotes) cout << h.info(curveInfo);
        if (options.povRayOutput)
        {
          cout << "#declare Order = " << options.traversalOrder << ";" << endl;
          cout << "#include \"" << options.povRayStyle << "\"" << endl;
          int size = (1 << (d * options.traversalOrder));
          int minCoord = 1 - (1 << options.traversalOrder);
          vector<FPt> pr;
          for (int i = 0; i < size; ++i) pr.push_back(convert(v[i],options.traversalOrder).povRay(options.povRayGap,options.traversalOrder));
          if (options.povRayContraction > 0)
          {
            vector<int> pred; 
            vector<int> succ; 
            int neighbour = -1; 
            for (int i = 0; i < size; ++i) 
            {
              if (v[i][2] == minCoord) neighbour = i;
              pred.push_back(neighbour); 
              succ.push_back(neighbour); // to be corrected in next loop
            }
            neighbour = size;
            for (int i = size-1; i >= 0; --i)
            {
              if (v[i][2] == minCoord) neighbour = i;
              if (neighbour < size) succ[i] = neighbour;
              if (pred[i] == -1) pred[i] = succ[i];
            }
            for (int i = 0; i < size; ++i)
            {
              // calculate destination on facet
              FPt fp;
              if (pred[i] == succ[i]) fp = pr[pred[i]];
              else fp = (pr[pred[i]] * (succ[i]-i) + pr[succ[i]] * (i-pred[i])) / (succ[i] - pred[i]);
              // calculate morph
              pr[i] = fp * options.povRayContraction + pr[i] * (1.0 - options.povRayContraction);
            }
          }
          double step = 1.0 / (double) (size - 1); // for colour gradient, if any
          cout << "Section(" << pr[0] << ", " << pr[0] << ", " << pr[1] << ", 0)" << endl;
          double progress = step;
          for (int i = 1; i < size-1; ++i, progress += step) 
            cout << "Section(" << pr[i-1] << ", " << pr[i] << ", " << pr[i+1] << ", " << progress << ")" << endl;
          cout << "Section(" << pr[size-2] << ", " << pr[size-1] << ", " << pr[size-1] << ", 1)" << endl;
        }
        v.clear();
        if (!options.countCurves && nrFound == options.maxOutput) break;
      }
      search.skip();
    }
    while (true);
    if (options.countCurves) cout << nrFound << (nrFound == 1 ? " curve" : " curves") << " found." << endl;
    if (options.batched) delete filter; else break;
  }
  while (true);
}

