/*******************************************************************
HILBEX
version 0.1, 1 Oct 2016

This file is part of the software package Hilbex by Herman
Haverkort. Use and distribution of this software is subject to
restrictions. Consult readme.txt for details, or download it from
http://spacefillingcurves.net/.
*******************************************************************/

const char* hex = "ILhbTCPEXJ4d7Z93";


Hilbert3DCurveSearch::Hilbert3DCurveSearch(Filter f, bool showProgress, bool onlyPatternIds):
  filter(f), progressIndicator(showProgress), patternIds(onlyPatternIds)
{
  partition = partitions.begin();
  transformation = transformations.begin();
  for (int i = 0; i < 2; ++i)
  {
    permutation[i] = permutations.begin();
    gateType[i] = gateTypes.begin();
    connections[i] = 0;
    reversals[i] = 0;
    reflections[i] = 0;
  }
  basePatternName = "Ca00";
}

bool Hilbert3DCurveSearch::symmetricBasePattern() const
{
  return (transformation->symbol <= 'o' && permutation[0]->symbol == permutation[1]->symbol);
}

bool Hilbert3DCurveSearch::vertexGated() const
{
  return (gateType[0]->vertexGate() && gateType[1]->vertexGate());
} 

bool Hilbert3DCurveSearch::vertexEdgeGated() const
{
  return ((gateType[0]->vertexGate() && gateType[1]->edgeGate()) || (gateType[0]->edgeGate() && gateType[1]->vertexGate()));
} 

bool Hilbert3DCurveSearch::vertexFacetGated() const
{
  return (gateType[0]->vertexGate() && gateType[1]->facetGate());
} 

bool Hilbert3DCurveSearch::edgeGated() const
{
  return (gateType[0]->edgeGate() && gateType[1]->edgeGate());
}

bool Hilbert3DCurveSearch::facetGated() const
{
  return (gateType[0]->facetGate() && gateType[1]->facetGate());
}



void Hilbert3DCurveSearch::calculateLocalTransformations()
{
  for (int i = 0; i < 2; ++i)
  {
    for (int k = 0; k < 4; ++k)
    {
      bool toReflect = (hexbit(reflections[i], k) == 1);
      int toReverse = hexbit(reversals[i], k);
      for (int tid = 0; tid < 48; ++tid)
      {
        GeometricTransformation t(tid);
        if (t.isReflection() != toReflect) continue;
        GPt entrGate = (GPt) localPattern[i][k] + t(gate[0]) / 2;
        GPt exitGate = (GPt) localPattern[i][k] + t(gate[8]) / 2;
        localTransformationPattern[i][k] = Transformation(t, 7*toReverse);
        if (toReverse == 0 && entrGate == localGate[i][k] && exitGate == localGate[i][k+1]) break;
        if (toReverse == 1 && exitGate == localGate[i][k] && entrGate == localGate[i][k+1]) break;
      }
    }  
  }
}


void Hilbert3DCurveSearch::calculateName(Hilbert3DCurve& h) const
{
  h.theName = basePatternName;
  h.theName += '.';
  h.theName += gateType[0]->symbol;
  h.theName += gateType[1]->symbol;

  // if it's the facet-gated curve, then this is it:
  if (facetGated()) return;
  
  // only vertex-gated and vertex-facet-gated curves use gate sequence encoding
  // only vertex-gated curves use reversal encoding
  if (vertexGated())
  {
    h.theName += '.';
    h.theName += hex[connections[0]]; 
    h.theName += hex[connections[1]];
    h.theName += '.';
    h.theName += hex[reversals[0]];
    h.theName += hex[reversals[1]];
  } 
  else if (vertexFacetGated())
  {
    h.theName += '.';
    h.theName += (connections[0] == 0 ? 'e' : 'f'); 
    h.theName += (connections[1] == 0 ? 'e' : 'f');
  }
  
  // all except facet-facet-gated curves use reflection encoding
  h.theName += '.';
  h.theName += hex[reflections[0]];
  h.theName += hex[reflections[1]];
}


void Hilbert3DCurveSearch::calculateDescription(Hilbert3DCurve& h)
{
  for (int i = 0; i < 4; ++i) 
  { 
    gate[i+1] = localGate[0][i+1];
    h.basePattern[i] =           localPattern[0][i];
    h.transformationPattern[i] = localTransformationPattern[0][i];
  }
  for (int i = 0; i < 4; ++i) 
  {
    gate[7-i] =                    transformation->geometry(localGate[1][i+1]);
    h.basePattern[7-i] =           transformation->geometry(localPattern[1][i]);
    h.transformationPattern[7-i] = (*transformation) * localTransformationPattern[1][i];
  }
}


bool Hilbert3DCurveSearch::findValidBasePattern()
{
  while (partition != partitions.end())
  {
    basePatternName[0] = partition->symbol;
    if (filter.name().length() > 0 && filter.name()[0] != partition->symbol) { ++partition; continue; }
    while (transformation != transformations.end())
    {
      basePatternName[1] = transformation->symbol;
      if (partition->transformations.find(transformation->symbol) == string::npos) { ++transformation; continue; }
      if (filter.name().length() > 1 && filter.name()[1] != transformation->symbol) { ++transformation; continue; }
      if (filter.requires(Symmetric) && transformation->symbol >= 'p') { ++transformation; continue; }
      while (permutation[0] != permutations.end())
      {
        basePatternName[2] = permutation[0]->symbol;
        if (partition->permutations.find(permutation[0]->symbol) == string::npos) { ++permutation[0]; continue; }
        if (filter.name().length() > 2 && filter.name()[2] != permutation[0]->symbol) { ++permutation[0]; continue; }
        while (permutation[1] != permutations.end())
        {
          basePatternName[3] = permutation[1]->symbol;
          if (filter.requires(WellFolded) && basePatternName != "Ca00") return false; // Ca00 would be the first to be generated; no need to look any further
          if (filter.forbids(WellFolded) && basePatternName == "Ca00") { ++permutation[1]; continue; }
          if (filter.name().length() > 3 && filter.name()[3] != permutation[1]->symbol) { ++permutation[1]; continue; }
          if (filter.requires(Symmetric) && permutation[1]->symbol != permutation[0]->symbol) { ++permutation[1]; continue; }
          if (partition->permutations.find(permutation[1]->symbol) == string::npos) { ++permutation[1]; continue; }
          if (transformation->symbol <= 'o' && permutation[1]->symbol < permutation[0]->symbol) { ++permutation[1]; continue; }
          for (int i = 0; i < 2; ++i) for (int k = 0; k < 4; ++k) localPattern[i][k] = partition->pattern[permutation[i]->inv(k)];
          localPattern[0][4] = transformation->geometry          (partition->pattern[permutation[1]->inv(3)]);
          localPattern[1][4] = transformation->geometry.inverse()(partition->pattern[permutation[0]->inv(3)]);
          return true;
        }
        permutation[1] = permutations.begin();
        ++permutation[0];
      }
      permutation[0] = permutations.begin();
      ++transformation;
    }
    transformation = transformations.begin();
    ++partition;
  }
  return false;
}

bool Hilbert3DCurveSearch::findValidGates()
{
  while (gateType[0] != gateTypes.end())
  {
    if (filter.name().length() > 5 && filter.name()[5] != gateType[0]->symbol) { ++gateType[0]; continue; }
    switch (gateType[0]->symbol)
    {
      case 'f': // no such curves: facet gates only occur in curves Ca00.gs and Cl00.cf.**.**
        ++gateType[0]; continue; 

      case 'g': // only occurs in curve Ca00.gs
        if (basePatternName.compare("Ca00") != 0) { ++gateType[0]; continue; }
        break;

      case 's': // no such curves: facet gates only occur in curves Ca00.gs and Cl00.cf.**.**
        ++gateType[0]; continue; 
    }
    while (gateType[1] != gateTypes.end())
    {
      if (filter.name().length() > 6 && filter.name()[6] != gateType[1]->symbol) { ++gateType[1]; continue; }
      if (filter.requires(Symmetric) && gateType[1]->symbol != gateType[0]->symbol) { ++gateType[1]; continue; }
      bool eligible = true;

      // calculate the actual coordinates of the gates
      int depth = ((gateType[0]->symbol == 'c' || gateType[1]->symbol == 'c') ? 6 : 4);
      for (int i = 0; i < 2; ++i)
        for (int j = 0; j < d; ++j)
          localGate[i][0][j] = localPattern[i][0][j] * (6 - (*gateType[i])[j] * depth);
      gate[0] = localGate[0][0];
      gate[8] = transformation->geometry(localGate[1][0]);
      
      if (vertexGated())
      {
        if (
          filter.forbids(VertexGated) || 
          filter.requires(Centred) || 
          filter.requires(Standing)
        ) { ++gateType[1]; continue; }
        if (filter.requires(FaceContinuous) || filter.requires(Hyperorthogonal))
          if (basePatternName != "Ca00" && basePatternName != "Si00") { ++gateType[1]; continue; }
        if (filter.forbids(FaceContinuous))
          if (basePatternName == "Ca00" || basePatternName == "Si00") { ++gateType[1]; continue; }
          
        if (collinear(localPattern[0][0], transformation->geometry(localPattern[1][0])))
        {
          eligible =
            !filter.forbids(EdgeCrossing) &&
            
          // first and last pairs of octants must share an edge
            collinear(localPattern[0][0], localPattern[0][1]) &&
            collinear(localPattern[1][1], localPattern[1][0]) &&
          
          // first and last triples of octants must share a facet
            coplanar(localPattern[0][0], localPattern[0][2]) &&
            coplanar(localPattern[1][2], localPattern[1][0]) &&

          // even gates have odd coordinate sum, so no connections through the centre at even gates
            coplanar(localPattern[0][1], localPattern[0][2]) &&
            coplanar(localPattern[0][3], localPattern[0][4]) &&
            coplanar(localPattern[1][1], localPattern[1][2]);
        }
        else if (coplanar(localPattern[0][0], transformation->geometry(localPattern[1][0]))) 
        {
          eligible = !filter.forbids(FacetCrossing);
          // all gates have odd coordinate sum, so no connections through the centre
          for (int i = 0; i < 2; ++i)
            for (int k = 1; eligible && k < 5; ++k)
              eligible = coplanar(localPattern[i][k-1], localPattern[i][k]);
        }
        else eligible = false;
      }
      else if (vertexEdgeGated())
      {
        if (
           filter.forbids(VertexEdgeGated) ||
           filter.forbids(FacetCrossing) ||
           filter.requires(OrderPreserving) ||
           filter.requires(FaceContinuous) ||
           filter.requires(Hyperorthogonal)
        ) { ++gateType[1]; continue; }
        
        // if base pattern symmetric, take lexicographically smallest gate ordering
        if (symmetricBasePattern() && gateType[0]->symbol != 'c') { ++gateType[1]; continue; }
      
        // ensure that the gates lie on a common facet but not on a common edge
        eligible = !collinear(gate[0], gate[8]) && coplanar(gate[0], gate[8]);
        
        // odd-numbered interior gates lie halfway on octant edges, so they cannot connect octants that do not share an interior edge
        for (int i = 0; i < 2; ++i)
          for (int k = 1; eligible && k < 4; k += 2)
            eligible = coplanar(localPattern[i][k-1], localPattern[i][k]);

        // the gates before and after a gate in the centre, must be halfway on octant edges on the outside of the cube, so
        // connections between octants that do not share an outer facet must be preceded and followed by connections between octants
        // that share an outer edge
        for (int i = 0; eligible && i < 2; ++i)
          if (!coplanar(localPattern[i][1], localPattern[i][2]))
            eligible = collinear(localPattern[i][0], localPattern[i][1]) && 
                       collinear(localPattern[i][2], localPattern[i][3]);
        if (eligible && !coplanar(localPattern[0][3], localPattern[0][4]))
          eligible = collinear(localPattern[0][2], localPattern[0][3]) && 
                     collinear(localPattern[1][2], localPattern[1][3]);
                     
        // the octant with the vertex gate and the previous/next octant must share an edge
        eligible = eligible &&
          (gateType[0]->symbol == 'c' ? collinear(localPattern[0][0], localPattern[0][1]) : collinear(localPattern[1][0], localPattern[1][1]));
      }
      else if (vertexFacetGated())
      {
        eligible = 
          !filter.forbids(VertexFacetGated) && 
          !filter.forbids(CubeCrossing) &&
          !filter.requires(OrderPreserving) &&
          !filter.requires(Centred) &&
          !filter.requires(FaceContinuous) &&
          !filter.requires(Hyperorthogonal) &&
          basePatternName.compare("Cl00") == 0 && 
          gateType[0]->symbol == 'c' && gateType[1]->symbol == 'f';
      }
      else if (edgeGated())
      {
        eligible = 
          !filter.forbids(EdgeGated) && 
          !filter.forbids(CubeCrossing) &&
          !filter.requires(Centred) &&
          !filter.requires(FaceContinuous) &&
          !filter.requires(Hyperorthogonal) &&
          basePatternName.compare("Cd00") == 0 && 
          gateType[0]->symbol == 'r' && (gateType[1]->symbol == 't' || gateType[1]->symbol == 'v');
      }
      else if (facetGated())
      {
        eligible = 
          !filter.forbids(FacetGated) && 
          !filter.forbids(CubeCrossing) &&
          !filter.requires(Centred) &&
          !filter.forbids(FaceContinuous) &&
          !filter.forbids(Hyperorthogonal) &&
          basePatternName.compare("Ca00") == 0 && 
          gateType[0]->symbol == 'g' && gateType[1]->symbol == 's';
      }
      else eligible = false;
      if (!eligible) { ++gateType[1]; continue; }
      return true;
    }
    gateType[1] = gateTypes.begin();
    ++gateType[0];
  }
  return false;
}

bool Hilbert3DCurveSearch::verifyGatesEdgeCrossingVertexGatedCurve(int i)
{
  // connections: first bit encodes g3 (0 if in the centre of the cube; 1 if on an edge)
  //              remaining three bits encode coordinates of g4 (0 if in the centre; 1 if not)
  
  // g1 is on edge shared by first two octants:
  localGate[i][1] = (GPt) (localPattern[i][0] + localPattern[i][1]);
  
  // g2 is on midpoint of facet shared by first three octants
  if (collinear(localPattern[i][1], localPattern[i][2]))
    localGate[i][2] = (GPt) (localPattern[i][0] + localPattern[i][2]);
  else
    localGate[i][2] = (GPt) (localPattern[i][1] + localPattern[i][2]);
  
  // g3 is either in the centre, or on the edge shared by the 3rd and the 4th octant
  // the latter is possible if and only if it is an edge of the facet that contains g2, so g1, g2, g3 and g4 are all on the same facet
  if (hexbit(connections[i], 0) == 0)
    localGate[i][3] = GPt(0,0,0);
  else
  {
    if (partition->symbol != 'C') return false;
    localGate[i][3] = (GPt) (localPattern[i][2] + localPattern[i][3]);
  }
  
  for (int j = 0; j < 3; ++j) localGate[i][4][j] = localPattern[i][3][j] * hexbit(connections[i], j+1) * 6;
  if (sqDistance(localGate[i][3], localGate[i][4]) != 36) return false;
  
  return true;
}

bool Hilbert3DCurveSearch::verifyGatesFacetCrossingVertexGatedCurve(int i)
{
  // connections: k-th bit encodes g_k: if k = 1 or g_{k-1}[2] = 0, store abs(g_k[2]), otherwise store abs(g_k[1]);

  // g0 is on outside corner of first octant;
  // all remaining gates are facet midpoints, so with two zero coordinates and one non-zero coordinate
  bool connected = false;
  for (int j = 0; !connected && j < 3; ++j) // try three candidates for the non-zero coordinate
  {
    localGate[i][1] = GPt(0,0,0); localGate[i][1][j] = localPattern[i][0][j] * 6;
    if (localGate[i][1][j] == localPattern[i][1][j]*6 && (localGate[i][1][2] != 0) == (hexbit(connections[i], 0) != 0)) 
      connected = true;
  }
  if (!connected) return false;

  for (int k = 2; k < 5; ++k)
  {
    connected = false;
    int encoded = (localGate[i][k-1][2] == 0 ? 2 : 1);
    bool nonzero = (hexbit(connections[i], k-1) != 0);
    for (int j = 0; !connected && j < 3; ++j) // try three candidates for the non-zero coordinate
    {
      if (localGate[i][k-1][j] != 0) continue; // cannot be the same as the previous gate
      localGate[i][k] = GPt(0,0,0); localGate[i][k][j] = localPattern[i][k-1][j] * 6;
      if (localGate[i][k][j] == localPattern[i][k][j]*6 && (localGate[i][k][encoded] != 0) == nonzero)
        connected = true;
    }
    if (!connected) return false;
  }
  
  return true;
}

bool Hilbert3DCurveSearch::verifyGatesVertexGatedCurve(int i)
{
  if (collinear(gate[0], gate[8])) return verifyGatesEdgeCrossingVertexGatedCurve(i);
  else return verifyGatesFacetCrossingVertexGatedCurve(i);
}

bool Hilbert3DCurveSearch::findValidVertexGatedCurve()
{
  while (connections[0] < 16)
  {
    if (filter.name().length() > 8 && filter.name()[8] != hex[connections[0]]) { ++connections[0]; continue; }
    if (!verifyGatesVertexGatedCurve(0)) { ++connections[0]; continue; }
    while (connections[1] < 16)
    {
      if (filter.name().length() > 9 && filter.name()[9] != hex[connections[1]]) { ++connections[1]; continue; }
      if (filter.requires(Symmetric) && connections[1] != connections[0]) { ++connections[1]; continue; } 

      // if base pattern is symmetric, use only the lexicographically smallest connection scheme
      if (symmetricBasePattern() && connections[0] > connections[1]) { ++connections[1]; continue; }
      
      if (!verifyGatesVertexGatedCurve(1)) { ++connections[1]; continue; }
      
      // verify that first half and second half connect
      if (localGate[0][4] != transformation->geometry(localGate[1][4])) { ++connections[1]; continue; }
      if (progressIndicator) cerr << basePatternName << ".cc." << hex[connections[0]] << hex[connections[1]] << "...\r";            
      
      while (reversals[0] < 16)
      {
        // symmetric curves have reversal coding 0000 0000; other order-preserving curves have reversal coding 0000 1111
        if ((filter.requires(OrderPreserving) || filter.requires(Symmetric)) && reversals[0] > 0) break;
        if (filter.name().length() > 11 && filter.name()[11] != hex[reversals[0]]) { ++reversals[0]; continue; }
        while (reversals[1] < 16)
        {
          // symmetric curves have reversal coding 0000 0000; other order-preserving curves have reversal coding 0000 1111
          if (filter.requires(Symmetric) && reversals[1] != 0) break;
          if (filter.requires(OrderPreserving) && reversals[1] != 0 && reversals[1] != 15) { ++reversals[1]; continue; }
          if (filter.forbids(OrderPreserving) && reversals[0] == 0 && reversals[1] == 15) break;
          if (filter.name().length() > 12 && filter.name()[12] != hex[reversals[1]]) { ++reversals[1]; continue; }
          // if connection scheme is symmetric, use only the lexicographically smallest name
          if (symmetricBasePattern() && connections[0] == connections[1] && reversals[0] > 15-reversals[1]) { ++reversals[1]; continue; }
          while (reflections[0] < 16)
          {
            if (filter.name().length() > 14 && filter.name()[14] != hex[reflections[0]]) { ++reflections[0]; continue; }
            while (reflections[1] < 16)
            {
              if (filter.name().length() > 15 && filter.name()[15] != hex[reflections[1]]) { ++reflections[1]; continue; }
              if ((filter.requires(Symmetric) || filter.requires(OrderPreserving) && reversals[1] == 0)
                && reflections[1] != reflections[0]) { ++reflections[1]; continue; }
              if (symmetricBasePattern() && connections[0] == connections[1])
              {
                // symmetric connection scheme
                bool symmetric;
                if (collinear(gate[0], gate[8]))
                {
                  // use only the lexicographically smallest name
                  if (reversals[0] == 15-reversals[1] && reflections[0] > 15-reflections[1]) { ++reflections[1]; continue; }
                  symmetric = ((reversals[0]^reflections[0]) == (reversals[1]^reflections[1]));
                }
                else
                { 
                  // use only the lexicographically smallest name
                  if (reversals[0] == 15-reversals[1] && reflections[0] > reflections[1]) { ++reflections[1]; continue; }
                  symmetric = (reflections[0] == reflections[1]);
                }
                
                // if curve is symmetric, reverse only the octants in the second half of the curve
                if (symmetric && (reversals[0] != 0 || reversals[1] != 0)) { ++reflections[1]; continue; }
                if (symmetric && filter.forbids(Symmetric)) { ++reflections[1]; continue; }
              }
              return true;
            }
            reflections[1] = 0;
            ++reflections[0];
          }
          reflections[0] = 0;
          ++reversals[1];
        }
        reversals[1] = 0;
        ++reversals[0];
      }
      reversals[0] = 0;
      ++connections[1];
    }
    connections[1] = 0;
    ++connections[0];
  }
  return false;            
}

bool Hilbert3DCurveSearch::findValidVertexFacetGatedCurve()
{
  // localGate[0][0] = GPt(-6,-6,-6);   localGate[1][0] = GPt(-6, 0, 0);
  localGate[0][1] = GPt( 0,-3,-3);   localGate[1][1] = GPt( 0,-3,-3);
  localGate[0][3] = GPt( 0, 3,-3);   localGate[1][3] = GPt( 0, 3,-3);
  localGate[0][4] = GPt(-6, 6, 0);   localGate[1][4] = GPt(-6, 6, 0);
  while (connections[0] < 2)
  {
    if (filter.name().length() > 8 && filter.name()[8] != (connections[0] == 0 ? 'e' : 'f')) { ++connections[0]; continue; }
    localGate[0][2] = GPt( 6, 0, -6 + 6*connections[0]);
    while (connections[1] < 2)
    {
      if (filter.name().length() > 9 && filter.name()[9] != (connections[1] == 0 ? 'e' : 'f')) { ++connections[1]; continue; }
      reversals[0] = 0x5;
      reversals[1] = 0x5;
      localGate[1][2] = GPt( 6, 0, -6 + 6*connections[1]);
      while (reflections[0] < 16)
      {
        if (filter.name().length() > 11 && filter.name()[11] != hex[reflections[0]]) { ++reflections[0]; continue; }
        while (reflections[1] < 16)
        {
          if (filter.name().length() > 12 && filter.name()[12] != hex[reflections[1]]) { ++reflections[1]; continue; }
          return true;
        }
        reflections[1] = 0;
        ++reflections[0];
      }
      reflections[0] = 0;
      ++connections[1];
    }
    connections[1] = 0;
    ++connections[0];
  }
  return false;
}

bool Hilbert3DCurveSearch::verifyGatesReflectionDeterminedCurve(int i)
{
  for (int k = 0; k < 4; ++k)
  {
    bool toReflect = hexbit(reflections[i], k);
    bool connected = false;
    for (int tid = 0; !connected && tid < 48; ++tid)
    {
      GeometricTransformation t(tid);
      if (t.isReflection() != toReflect) continue;
      GPt entrGate = (GPt) localPattern[i][k] + t(gate[0]) / 2;
      GPt exitGate = (GPt) localPattern[i][k] + t(gate[8]) / 2;
      int direction;
      if (entrGate == localGate[i][k]) 
      {
        localGate[i][k+1] = exitGate;
        direction = 0;
      }
      else if (exitGate == localGate[i][k]) 
      { 
        localGate[i][k+1] = entrGate;
        direction = 7;
      }
      else continue;
      connected = (sqDistance(localGate[i][k+1], (GPt) localPattern[i][k+1]) <= 27);
      localTransformationPattern[i][k] = Transformation(t, direction);
    }
    if (!connected) return false;
  }
  return true;
}

bool Hilbert3DCurveSearch::findValidReflectionDeterminedCurve()
{
  while (reflections[0] < 16)
  {
    if (filter.name().length() > 8 && filter.name()[8] != hex[reflections[0]]) { ++reflections[0]; continue; }
    if (!verifyGatesReflectionDeterminedCurve(0)) { ++reflections[0]; continue; }
    if (filter.requires(Centred) && localGate[0][4] != GPt(0,0,0) ) { ++reflections[0]; continue; }
    if (filter.forbids(Centred) && localGate[0][4] == GPt(0,0,0) ) { ++reflections[0]; continue; }
    while (reflections[1] < 16)
    {
      if (filter.name().length() > 9 && filter.name()[9] != hex[reflections[1]]) { ++reflections[1]; continue; }
      if (!verifyGatesReflectionDeterminedCurve(1)) { ++reflections[1]; continue; }
      
      // verify that first half and second half connect
      if (localGate[0][4] != transformation->geometry(localGate[1][4])) { ++reflections[1]; continue; }
      
      return true;
    }
    reflections[1] = 0;
    ++reflections[0];
  }
  return false;            
}

bool Hilbert3DCurveSearch::findValidCurve(Hilbert3DCurve& h)
{
  if (!filter.consistent()) return false;
  while (findValidBasePattern())
  {
    if (progressIndicator) cerr << basePatternName << "...      \r";
    if (patternIds) for (int dir = 0; dir < 2; ++dir)
    {
      OPt pattern[8];
      int directionMask = dir * 7;
      for (int i = 0; i < 4; ++i) pattern[directionMask^i  ] = localPattern[0][i];
      for (int i = 0; i < 4; ++i) pattern[directionMask^i^7] = transformation->geometry(localPattern[1][i]);
      
      // put into canonical form
      for (int i = 7; i >= 0; --i) 
        for (int j = 0; j < d; ++j) 
          pattern[i][j] = pattern[i][j] * -pattern[0][j];
      OPt neighbour[3]; // neighbour[j] will be the octant that differs from the starting octant only in coordinate j
 
      for (int j = 0; j < d; ++j) { neighbour[j] = pattern[0]; neighbour[j][j] = 1; } 
      int rank[3]; // rank[j] is going to contain the position in the pattern of neighbour[j]
      for (int i = 1; i < 8; ++i) for (int j = 0; j < d; ++j) if (pattern[i] == neighbour[j]) rank[j] = i;
      if (rank[0] < rank[1]) { swap(rank[0], rank[1]); for (int i = 0; i < 8; ++i) swap(pattern[i][0], pattern[i][1]); }
      if (rank[1] < rank[2]) { swap(rank[1], rank[2]); for (int i = 0; i < 8; ++i) swap(pattern[i][1], pattern[i][2]); }
      if (rank[0] < rank[1]) { swap(rank[0], rank[1]); for (int i = 0; i < 8; ++i) swap(pattern[i][0], pattern[i][1]); }
      
      // determine r5 to r8 (indexed 0 to 3)
      int r[4]; for (int i = 0; i < 4; ++i) r[i] = 8;
      for (int i = 1; i < 8; ++i)
      {
        if (pattern[i] == OPt(-1, 1, 1)) r[0] = i - (r[1]<8) - (r[2]<8) - (r[3]<8); else
        if (pattern[i] == OPt( 1,-1, 1)) r[1] = i            - (r[2]<8) - (r[3]<8); else
        if (pattern[i] == OPt( 1, 1,-1)) r[2] = i                       - (r[3]<8); else
        if (pattern[i] == OPt( 1, 1, 1)) r[3] = i;
      }
      
      cout << ((((r[0]-1)*5 + (r[1]-1))*6 + (r[2]-1))*7 + (r[3]-1)) << endl;        
      
      // if symmetric, that is it
      if (basePatternName[1] <= 'o' && basePatternName[2] == basePatternName[3]) break;
    }
    else while (findValidGates())
    {
      bool found = false;
      
      if (vertexGated())
      { if (findValidVertexGatedCurve()) { calculateLocalTransformations(); found = true; } }
      else if (vertexEdgeGated() || edgeGated() || facetGated())
      { if (findValidReflectionDeterminedCurve()) { found = true; } }
      else if (vertexFacetGated())
      { if (findValidVertexFacetGatedCurve()) { calculateLocalTransformations(); found = true; } }
      
      if (found)
      {
        calculateDescription(h);
        calculateName(h);
        return true;
      }
      
      // proceed to next gate combination
      connections[0] = 0; connections[1] = 0;
      reversals[0] = 0; reversals[1] = 0;
      reflections[0] = 0; reflections[1] = 0;
      ++gateType[1];
    }
    // proceed to next base pattern
    gateType[0] = gateTypes.begin();
    ++permutation[1];
  }
  return false;
}


bool Hilbert3DCurveSearch::skip()
{
  ++reflections[1];
}

