/*******************************************************************
HILBEX
version 0.1, 1 Oct 2016

This file is part of the software package Hilbex by Herman
Haverkort. Use and distribution of this software is subject to
restrictions. Consult readme.txt for details, or download it from
http://spacefillingcurves.net/.
*******************************************************************/

#include <algorithm>
using std::swap;

string rank(const int i)
{
  stringstream s;
  s << i;
  if (i / 10 == 1) s << "th";
  else switch(i % 10)
  {
    case 1:  s << "st"; break;
    case 2:  s << "nd"; break;
    case 3:  s << "rd"; break;
    default: s << "th"; break;
  }
  return s.str();
}


RecursiveTraversal::RecursiveTraversal()
{
  // non-crashing default: Z-order
  for (int i = 0; i < 8; ++i)
  {  
    basePattern[i] = convert(NPt(i >> 2 & 1, i >> 1 & 1, i & 1));
    invBasePattern[i >> 2 & 1][i >> 1 & 1][i & 1] = i;
  }
}


RecursiveTraversal::RecursiveTraversal(istream& s)
{
  char c; s >> std::ws; c = s.peek();
  if (!s.good() || c != '<') 
  {
    cerr << "Error parsing traversal: missing description" << endl;
    exit(1);
  }
  s >> c; 
  basePattern[0] = OPt(-1,-1,-1);
  for (int i = 0; i < 7; ++i)
  {
    transformationPattern[i] = Transformation(s);
    basePattern[i+1] = basePattern[i];
    do
    {  
      s >> std::ws; c = s.peek();
      if (!s.good())
      {
        cerr << "Incomplete traversal description" << endl;
        exit(1);
      }
      if (c == '[' || c == '{') break;
      int step; s >> step;
      int sign = (step < 0 ? -1 : 1);
      step = step * sign - 1;
      if (basePattern[i+1][step] == sign)
      {
        // flip the pattern so far to the other side so that the step can be made in the desired direction
        // (if positive and negative steps do not alternate, strange things may happen, but no crash)
        for (int j = 0; j <= i; ++j) basePattern[j][step] = -basePattern[j][step];
      }
      basePattern[i+1][step] = sign;
    }
    while (true);
  }
  transformationPattern[7] = Transformation(s);
  s >> std::ws >> c;
  if (c != '>')
  {
    cerr << "Missing closing > of traversal description" << endl;
    exit(1);
  }
  for (int i = 0; i < 8; ++i) 
    invBasePattern[i >> 2 & 1][i >> 1 & 1][i & 1] = -1;
  for (int i = 0; i < 8; ++i)
    invBasePattern[convert(basePattern[i])[0]][convert(basePattern[i])[1]][convert(basePattern[i])[2]] = i;
  for (int i = 0; i < 8; ++i) 
    if (invBasePattern[i >> 2 & 1][i >> 1 & 1][i & 1] == -1)
    {
      cerr << "Invalid base pattern in traversal description" << endl;
      for (int i = 0; i < 8; ++i) cerr << basePattern[i]; cerr << endl;
      exit(1);
    }
}


string RecursiveTraversal::description(const Transformation& t) const
{
  OPt base[8];
  for (int i = 0; i < 8; ++i) base[i] = t.geometry(basePattern[i^t.direction]);
  stringstream s;
  s << "< ";
  for (int i = 0; i < 8; ++i)
  {
    s << t * transformationPattern[i^t.direction] * t.inverse();
    if (i == 7) break;
    for (int j = 0; j < d; ++j)
      if (base[i+1][j] < base[i][j]) s << ' ' << -(j+1);
      else if (base[i+1][j] > base[i][j]) s << ' ' << (j+1);
    s << ' ';
  }
  s << " >";
  return s.str();
}


void RecursiveTraversal::generateTraversal(
  vector<OPt>& v, int depth, const Transformation& t, const OPt offset) 
const
{
  if (depth <= 0) 
  {
    v.push_back(offset);
    return;
  }
  --depth;
  int size = (1 << depth);
  for (int i = 0; i < 8; ++i)
    generateTraversal(
      v,
      depth,
      t * transformationPattern[i^t.direction],
      offset + t.geometry(basePattern[i^t.direction]) * size
    );
}

TransformationCandidates::TransformationCandidates()
{
  count[0] = 48; count[1] = 48;
}

int TransformationCandidates::nrCandidates() const
{
  return count[0] + count[1];
}

int TransformationCandidates::nrCandidates(const int direction) const
{
  return count[direction == 0 ? 0 : 1];
}

Transformation TransformationCandidates::candidate(const int direction) const
{
  if (direction == 0) return Transformation(geometry[0], 0);
  else                return Transformation(geometry[1], 7);
}

void LimitedTraversal::calculateCanonicalForm(const Transformation& embedding, const Transformation& symmetry, vector<Approximant>& canonical, OPt localGate[2][5])
{
  // apply inverse of embedding to compute traversal in canonical form
  if (canonical.empty())
  {
    for (int k = 0; k < a.size(); ++k)
    {
      canonical.push_back(Approximant());
      for (int i = 0; i < a[k].size(); ++i) canonical.back().push_back(OPt(0,0,0));
    }
  }
   
  for (int k = 0; k < a.size(); ++k)
  {
    int directionMask = (embedding.direction == 0 ? 0 : a[k].size() - 1);
    for (int i = 0; i < a[k].size(); ++i) canonical[k][i] = embedding.geometry.inverse()(a[k][i^directionMask]);
  }
    
  // check for self-similarity and try to determine transformations in the octants  
  for (int o = 0; o < 8; ++o)      
  {
    for (int dir = 0; dir < 2; ++dir)
    {
      candidates[o].count[dir] = 0;
      int directionMask = (dir == 0 ? 0 : 7);
      for (int tid = 0; tid < 48; ++tid)
      {
        candidates[o].geometry[dir] = GeometricTransformation(tid);
        bool match = true;
        OPt octantCentre = canonical[1][o] * 2;
        for (int i = 0, j = o * 8; match && i < 8; ++i, ++j)
          match = (octantCentre + candidates[o].geometry[dir](canonical[1][i^directionMask]) == canonical[2][j]);
        if (match) 
        {
          candidates[o].count[dir] = 1;
          break;
        }
      }
    }
    if (candidates[o].nrCandidates() == 0)
    {
      cout << "Traversal is not octant-wise self-similar: " << rank(o+1) << " octant of 2nd-order curve is not similar to 1st-order curve." <<endl;
      properties.confirm(Invalid);
      return;
    }
  }
  
  // gate configuration (gate types)
  for (int i = 0; i < 5; ++i) localGate[0][i] = canonical[2][i*8];
  for (int i = 0; i < 5; ++i) localGate[1][i] = symmetry.geometry.inverse()(canonical[2][63-i*8]);
}


void LimitedTraversal::analyseProperties() 
{
  // all analysis requires at least A1 to be specified, so if only A0 is there (a single vertex), we're done
  if (a.size() < 2) return;

  // calculate partition directly
  int pairsOnEdge = 0; int pairsOnFacet = 0;
  for (int i = 0; i < 3; ++i)
    for (int j = i+1; j < 4; ++j)
    {
      if (coplanar(a[1][i], a[1][j])) ++pairsOnFacet;
      if (collinear(a[1][i], a[1][j])) ++pairsOnEdge;
    }
  if (pairsOnEdge == 4 && pairsOnFacet == 6) prefix = "C???"; else
  if (pairsOnEdge == 3 && pairsOnFacet == 6) prefix = "Y???"; else
  if (pairsOnEdge == 3 && pairsOnFacet == 5) prefix = "S???"; else
  if (pairsOnEdge == 2 && pairsOnFacet == 5) prefix = "L???"; else
  if (pairsOnEdge == 2 && pairsOnFacet == 4) prefix = "N???"; else
  if (pairsOnEdge == 0 && pairsOnFacet == 6) prefix = "X???"; else
  { 
    cerr << "Error in traversal: 1st-order curve starts with " << a[1][0] << a[1][1] << a[1][2] << a[1][3] << "." << endl; 
    exit(1); 
  }
  vector<Partition>::const_iterator partition = partitions.begin();
  while (partition->symbol != prefix[0]) ++partition;

  // identify base pattern
  Transformation embedding;
  vector<Symmetry>::iterator symmetry;
  bool match = false;
  for (int tid = 0; !match && tid < 96; ++tid) // try all possible transformations of the whole curve
  {
    embedding = Transformation(tid);
    vector<Permutation>::const_iterator permutation[2];
    for (permutation[0] = permutations.begin(); !match && permutation[0] != permutations.end(); ++permutation[0])
    {
      if (partition->permutations.find(permutation[0]->symbol) == string::npos) continue;
      match = true;
      for (int i = 0; match && i < 4; ++i)
        match = (embedding.geometry(partition->pattern[permutation[0]->inv(i)]) == a[1][i^embedding.direction]);
      if (!match) continue;
      match = false;
      prefix[2] = permutation[0]->symbol;
      Symmetries::const_iterator transformation;
      for (symmetry = transformations.begin(); symmetry != transformations.end(); ++symmetry)
      {
        if (partition->transformations.find(symmetry->symbol) == string::npos) continue;
        prefix[1] = symmetry->symbol;
        for (permutation[1] = permutations.begin(); !match && permutation[1] != permutations.end(); ++permutation[1])
        {
          if (partition->permutations.find(permutation[1]->symbol) == string::npos) continue;
          prefix[3] = permutation[1]->symbol;
          if (symmetry->symbol <= 'o' && permutation[1]->symbol < permutation[0]->symbol) continue;
          match = true;
          for (int i = 4; match && i < 8; ++i)
            match = (embedding.geometry(symmetry->geometry(partition->pattern[permutation[1]->inv(i^7)])) == a[1][i^embedding.direction]);
        }
        if (match) break; // not in loop guard, because that messes up symmetry
      }
    }
  }
  
  // check for properties that can be ruled out efficiently:
  // face-continuity and hyperorthogonality:
  bool faceContinuous = true;
  bool hyperorthogonal = true;
  for (int k = 1; faceContinuous && k < a.size(); ++k)
  {
    faceContinuous = (sqDistance(a[k][0], a[k][1]) == 4);
    for (int i = 2; faceContinuous && i < a[k].size(); ++i)
    {
      faceContinuous = (sqDistance(a[k][i-1], a[k][i]) == 4);
      hyperorthogonal = hyperorthogonal && (sqDistance(a[k][i-2], a[k][i]) == 8);
    }
  }
  if (!faceContinuous) properties.reject(FaceContinuous);
  if (!faceContinuous || !hyperorthogonal) properties.reject(Hyperorthogonal);

  // gate configuration (crossing type)
  if (!collinear(a.back().front(), a.back().back())) 
  {
    properties.reject(EdgeCrossing);
    if (!coplanar(a.back().front(), a.back().back())) 
    {
      properties.reject(FacetCrossing);
      properties.confirm(CubeCrossing);
    }
  }

  // the remaining tests are meaningful only if at least A2 is there:
  if (a.size() < 3) return;

  // verify continuity
  for (int i = 1; i < a.back().size(); ++i)
    if (sqDistance(a.back()[i-1], a.back()[i]) > 12)
    {
      cout << "Traversal is discontinuous: it jumps from " << convert(a.back()[i-1], a.size()) << " to " << convert(a.back()[i], a.size()) << "." << endl;
      properties.confirm(Invalid);
      return;
    }

  // calculcate canonical form, local gate coordinates, and candidate transformations in the octants
  vector<Approximant> canonical;
  OPt localGate[2][5];
  calculateCanonicalForm(embedding, *symmetry, canonical, localGate);
  if (properties.is(Invalid)) return;

  prefix += '.';
  for (int i = 0; i < 2; ++i)
  {
    NPt gateType;
    for (int j = 0; j < d; ++j)
      gateType[j] = (localGate[i][0][j]*localGate[i][0][j] == 1 ? 1 : 0);
    switch(gateType[0] | gateType[1] << 1 | gateType[2] << 2)
    {
      case 0: prefix += 'c'; break;
      case 1: prefix += 'r'; break;
      case 2: prefix += 'v'; break;
      case 3: prefix += 's'; break;
      case 4: prefix += 't'; break;
      case 5: prefix += 'g'; break;
      case 6: prefix += 'f'; break;
      case 7: 
        cout << "Traversal cannot be octant-wise self-similar and continuous: it starts or ends in the interior of the cube." << endl;
        properties.confirm(Invalid);
        return;
    }
  }

  // if base pattern symmetric, pick lexicographically smallest order; 
  if (prefix[1] <= 'o' && prefix[2] == prefix[3] && prefix[5] > prefix[6])
  {
    swap(prefix[5], prefix[6]);
    embedding = embedding * (*symmetry);
    calculateCanonicalForm(embedding, *symmetry, canonical, localGate);
  }
  
  // centredness:
  bool centred = true;  
  for (int k = 1; centred && k < canonical.size(); ++k)
  {
    if (sqDistance(canonical[k][canonical[k].size()/2-1], OPt(0,0,0)) > 3) centred = false;
    else if (sqDistance(canonical[k][canonical[k].size()/2], OPt(0,0,0)) > 3) centred = false;
  }
  if (!centred) properties.reject(Centred);

  // order-preservation (if ruled out, than so is symmetry):
  bool orderPreserving = true;
  for (int o = 0; orderPreserving && o < 8; ++o)
  {
    orderPreserving = candidates[o].nrCandidates(0);
  }
  if (!orderPreserving) 
  {
    properties.reject(OrderPreserving);
    properties.reject(Symmetric);
  }
  
  // identify connection scheme / gate sequence
  if (prefix[5] == 'c' && prefix[6] == 'c')
  {
    int encoding[2];
    for (int i = 0; i < 2; ++i)
    {
      encoding[i] = 0;
      if (collinear(canonical[1][0], canonical[1][7]))
      {
        encoding[i] <<= 1; encoding[i] |= (sqDistance(localGate[i][3], OPt(0,0,0)) != 3);
        for (int j = 0; j < d; ++j)
        {
          encoding[i] <<= 1; encoding[i] |= (localGate[i][4][j]*localGate[i][4][j] > 1);
        }
      }
      else
      {
        encoding[i] <<= 1; encoding[i] |= (localGate[i][1][2]*localGate[i][1][2] > 1);
        for (int k = 2; k < 5; ++k)
        {
          int encoded = (localGate[i][k-1][2]*localGate[i][k-1][2] == 1 ? 2 : 1);
          encoding[i] <<= 1; encoding[i] |= (localGate[i][k][encoded]*localGate[i][k][encoded] > 1);
        }
      }
    }

    bool symmetricBase = (prefix[1] <= 'o' && prefix[2] == prefix[3]);        
    if (symmetricBase)
    {
      if (encoding[0] > encoding[1])
      {
        swap(encoding[0], encoding[1]);
        embedding = embedding * (*symmetry);
        calculateCanonicalForm(embedding, *symmetry, canonical, localGate);
      }
    }
    prefix += '.';
    prefix += hex[encoding[0]];
    prefix += hex[encoding[1]];

    bool fullySymmetric = (symmetricBase && prefix[8] == prefix[9]); // tentatively, now check octants

    for (int i = 0; fullySymmetric && i < 4; ++i)
      fullySymmetric = (candidates[i].nrCandidates(0) == 1 && candidates[i^7].nrCandidates(1) == 1 &&
        candidates[i].geometry[0] == symmetry->geometry * candidates[i^7].geometry[1]);

    // if base-pattern is asymmetric, or 
    // if fully symmetric order-preserving solution exists, 
    // then generate it
    
    if (!symmetricBase || fullySymmetric)
    {
      prefix += '.';
      int reversals[2];
      int reflections[2];
      for (int i = 0; i < 2; ++i)
      {
        reversals[i] = 0;
        reflections[i] = 0;
        for (int k = 0; k < 4; ++k)
        {
          int o = k ^ (7*i);
          int dir = ( candidates[o].nrCandidates(i) == 1 ? i : 1-i );
          reversals[i] <<= 1; reversals[i] |= dir ^ i;                
          reflections[i] <<= 1; reflections[i] |= candidates[o].geometry[dir].isReflection();   
          if (i == 1 && symmetry->geometry.isReflection()) reflections[i] ^= 1;
        }        
      }
      prefix += hex[reversals[0]]; prefix += hex[reversals[1]];
      prefix += '.';
      prefix += hex[reflections[0]]; prefix += hex[reflections[1]]; 
    }
  }
  else if (prefix[5] == 'c' && prefix[6] == 'f')
  {
    prefix += '.';
    prefix += (sqDistance(localGate[0][2], OPt(0,0,0)) == 11 ? 'f' : 'e');
    prefix += (sqDistance(localGate[1][2], OPt(0,0,0)) == 11 ? 'f' : 'e');
  }

  if (!(prefix[5] == 'c' && prefix[6] == 'c') && !(prefix[5] == 'g' && prefix[6] == 's'))
  {
    // work out reflections for vertex-edge, vertex-facet, or edge-gated curves
    // for vertex-edge and vertex-facet-gated directions alternate, starting with forward at the vertex-gated end
    // for edge-gated curves directions are harder to determine, but in case an octant of A2 allows two options, they will agree on the reflection
    prefix += '.';
    for (int i = 0; i < 2; ++i)
    {
      int reflections = 0;
      for (int k = 0; k < 4; ++k)
      {
        int o = k ^ (7*i);
        int dir = (sqDistance(localGate[i][0], OPt(0,0,0)) < 27) ^ (k & 1);  // correct direction for vertex-edge and vertex-facet gated curves
        if (candidates[o].nrCandidates(dir) == 0) dir = 1 - dir;             // for edge-gated curves and malconstructed curves
        reflections <<= 1; reflections |= candidates[o].geometry[dir].isReflection();
        if (i == 1 && symmetry->geometry.isReflection()) reflections ^= 1;
      }
      prefix += hex[reflections];
    }      
  }
  
  // the remaining tests are meaningful only if at least A3 is there:
  if (a.size() < 4) return;

  // check whether candidate transformations map canonical[k] to each octant of canonical[k+1]
  for (int k = 2; k < canonical.size()-1; ++k)
  {
    for (int o = 0; o < 8; ++o)      
    {
      for (int dir = 0; dir < 2; ++dir)
      {
        if (candidates[o].nrCandidates(dir) == 0) continue;
        int directionMask = (dir == 0 ? 0 : canonical[k].size()-1);
        bool match = true;
        OPt octantCentre = canonical[1][o] * (1 << k);
        for (int i = 0, j = o * canonical[k].size(); match && i < canonical[k].size(); ++i, ++j)
          match = (octantCentre + candidates[o].geometry[dir](canonical[k][i^directionMask]) == canonical[k+1][j]);
        if (!match) candidates[o].count[dir] = 0;
      }
      if (candidates[o].nrCandidates() == 0)
      {
        cout << "Traversal is not octant-wise self-similar: " << rank(o+1) << " octant of " << rank(k+1) << "-order curve is not similar to " << rank(k) << "-order curve." << endl;
        properties.confirm(Invalid);
        return;
      }
    }
  }

  // work out reversals and reflections for vertex-gated curves, if this could not be done with A2:
      
  if (prefix[5] == 'c' && prefix[6] == 'c' && prefix.length() < 16)
  {
    int reversals[2];
    int reflections[2];
    for (int i = 0; i < 2; ++i)
    {
      reversals[i] = 0;
      reflections[i] = 0;
      for (int k = 0; k < 4; ++k)
      {
        int o = k ^ (7*i);
        int dir = ( candidates[o].nrCandidates(i) == 1 ? i : 1-i ); 
        reversals[i] <<= 1; reversals[i] |= dir^i;                
        reflections[i] <<= 1; reflections[i] |= candidates[o].geometry[dir].isReflection();   
        if (i == 1 && symmetry->geometry.isReflection()) reflections[i] ^= 1;
      }
    }

    // make sure the lexicographically smallest solution is chosen  
    if (prefix[8] == prefix[9])
    {
      bool needToSwap;
      if (reversals[0] < 15-reversals[1])
        needToSwap = false;
      else if (reversals[0] > 15-reversals[1]) 
        needToSwap = true;
      else // reversals[0] == 15-reversals[1]
      {
        if (collinear(canonical[1][0], canonical[1][7])) 
          needToSwap = (reflections[0] > 15-reflections[1]);
        else
          needToSwap = (reflections[0] > reflections[1]);
      }
      if (needToSwap)
      {
        swap(reversals[0], reversals[1]);
        reversals[0] ^= 15; reversals[1] ^= 15;
        swap(reflections[0], reflections[1]);
        if (collinear(canonical[1][0], canonical[1][7])) { reflections[0] ^= 15; reflections[1] ^= 15; }
        // embedding = embedding * (*symmetry);        
        // do not bother actually updating embedding and recomputing canonical and localGate, since they will not be used anymore
      }
    }

    prefix += '.';
    prefix += hex[reversals[0]]; prefix += hex[reversals[1]];
    prefix += '.';
    prefix += hex[reflections[0]]; prefix += hex[reflections[1]]; 
  }
}

LimitedTraversal::LimitedTraversal()
{
  a.push_back(Approximant()); a.back().push_back(OPt(0,0,0));
}

LimitedTraversal::LimitedTraversal(istream& s)
{
  char c; s >> std::ws; c = s.peek();
  if (!s.good() || (c != '(' && c != '<'))
  {
    cerr << "Error parsing traversal: missing description" << endl;
    exit(1);
  }
  
  if (c == '<') // definition of recursive traversal
  {
    RecursiveTraversal r(s);
    for (int k = 0; k <= 3; ++k)
    {
      a.push_back(Approximant());
      r.generateTraversal(a.back(), k);
    }
    analyseProperties();
    return; 
  }
  
  // otherwerwise it must be a sequence of points
  vector<NPt> v;
  while (s.good() && c == '(')
  {
    int x0, x1, x2; char c1, c2, c3; bool parsingError = false;
    s >> c >> std::ws >> x0 >> std::ws >> c1 >> std::ws >> x1 >> std::ws >> c2 >> std::ws >> x2 >> std::ws >> c3;
    if (!s.good() || c1 != ',' || c2 != ',' || c3 != ')') parsingError = true;
    if (parsingError) 
    {
      cerr << "Error parsing traversal: cannot parse point " << c << x0 << c1 << x1 << c2 << x2 << c3 << endl;
      cerr << "Points read successfully: "; 
      if (v.size() <= 8) for (int i = 0; i < v.size(); ++i) cerr << v[i];
      else { for (int i = 0; i < 4; ++i) cerr << v[i]; cerr << "..."; for (int i = v.size()-4; i < v.size(); ++i) cerr << v[i]; }
      exit(1);
    }
    v.push_back(NPt(x0, x1, x2));
    s >> std::ws; c = s.peek();    
  }
  if (v.size() == 0) v.push_back(NPt(0,0,0));

  // determine order of the approximating curve just read
  int order = 0; while (v.size() > (1 << (d*order))) ++order;
  if (v.size() != (1 << d*order))
  {
    cerr << "Error in traversal: read " << v.size() << " points but expected a power of eight." << endl;
    exit(1);
  }

  // check whether all coordinates are in range
  for (int i = 0; i < v.size(); ++i)
    for (int j = 0; j < d; ++j)
      if (v[i][j] < 0 || v[i][j] >= (1 << order))
      {
        cerr << "Error in traversal: " << rank(i+1) << " point has coordinates " << v[i] << "," << endl
             << "but coordinates should be in range 0,...," << (1 << order) - 1 << "." << endl;
        exit(1);
      }
      
  // check consistency with octant-based traversal
  int groupMask = ~0;
  for (int k = 1; k < order; ++k)
  {
    groupMask <<= 3;
    for (int i = 0; i < v.size(); ++i)
      for (int j = 0; j < d; ++j)
        if ((v[i][j] >> k) != v[i&groupMask][j] >> k)
        {
          cout << "Traversal is not octant-based: " << rank((i&groupMask)+1) << " to " << rank((i&groupMask)+~groupMask+1) << " points "
               << "do not all lie in the same " << (~groupMask+1) << 'x' << (~groupMask+1) << 'x' << (~groupMask+1) << " cube." << endl;
          properties.confirm(Invalid);
          return;
        }
  }
       
  // now construct all levels
  for (int k = order; k >= 0; --k)
  {
    int blockVolume = 1 << (3 * k);
    a.push_back(Approximant());
    for (int i = 0; i < v.size(); i += blockVolume)
      a.back().push_back(convert(NPt(v[i][0] >> k, v[i][1] >> k, v[i][2] >> k), order - k));
  }
  
  analyseProperties();
}

bool LimitedTraversal::match(const RecursiveTraversal& r, Transformation& t) const
{
  vector<OPt> v;
  r.generateTraversal(v, a.size() - 1);
  bool match = false;
  for (int tid = 0; !match && tid < 96; ++tid)
  {
    t = Transformation(tid);
    match = true;
    int directionMask = (t.direction == 0 ? 0 : v.size() - 1);
    for (int i = 0; match && i < v.size(); ++i)
      match = (a.back()[i] == t.geometry(v[i^directionMask]));
  }  
  return match;
}

bool LimitedTraversal::requires(const Property p) const
{
  return properties.is(p);
}

bool LimitedTraversal::forbids(const Property p) const
{
  return properties.isNot(p);
}

string LimitedTraversal::name() const
{
  return prefix;
}

const TransformationCandidates& LimitedTraversal::transformation(const int i) const
{
  return candidates[i];
}

ostream& operator<<(ostream& o, const LimitedTraversal& t)
{
  int order = t.a.size()+1;
  for (LimitedTraversal::Approximant::const_iterator i = t.a.back().begin(); i != t.a.back().end(); ++i) o << convert(*i, order);
}

